# coding=utf-8
# Copyright 2019 The Tensor2Tensor Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Autoencoders."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensor2tensor.layers import common_attention
from tensor2tensor.layers import common_hparams
from tensor2tensor.layers import common_layers
from tensor2tensor.layers import discretization
from tensor2tensor.layers import latent_layers
from tensor2tensor.layers import modalities
from tensor2tensor.utils import t2t_model
from tensor2tensor.models.research import autoencoders

from galaxy2galaxy.utils import registry

import tensorflow as tf

def pack_images(images, rows, cols):
    """Helper utility to make a field of images."""
    shape = tf.shape(images)
    width = shape[-3]
    height = shape[-2]
    depth = shape[-1]
    images = tf.reshape(images, (-1, width, height, depth))
    batch = tf.shape(images)[0]
    rows = tf.minimum(rows, batch)
    cols = tf.minimum(batch // rows, cols)
    images = images[:rows * cols]
    images = tf.reshape(images, (rows, cols, width, height, depth))
    images = tf.transpose(images, [0, 2, 1, 3, 4])
    images = tf.reshape(images, [1, rows * width, cols * height, depth])
    return images

@registry.register_model
class ContinuousAutoencoderBasic(autoencoders.AutoencoderBasic):
  """Continuous version of the basic Autoencoder"""

  def reconstruction_loss(self, values, targets):
    hparams = self.hparams
    pz = tf.reduce_sum(tf.abs(values - targets)**2, axis=[-1, -2, -3]) / hparams.reconstruction_loss_sigma**2
    return tf.reduce_mean(pz)

  def image_summary(self, name, image_logits, max_outputs=1, rows=8, cols=8):
    """Helper for image summaries that are safe on TPU."""
    if len(image_logits.get_shape()) != 4:
      tf.logging.info("Not generating image summary, maybe not an image.")
      return
    return tf.summary.image(
        name, pack_images(image_logits, rows, cols),
        #common_layers.tpu_safe_image_summary(pack_images(tensor, rows, cols)),
        max_outputs=max_outputs)

  def body(self, features):
    hparams = self.hparams
    is_training = hparams.mode == tf.estimator.ModeKeys.TRAIN
    encoder_layers = None
    self.is1d = hparams.sample_width == 1
    if (hparams.mode != tf.estimator.ModeKeys.PREDICT
        or self._encode_on_predict):
      labels = features["targets_raw"]
      labels_shape = common_layers.shape_list(labels)
      # handle videos
      if len(labels.shape) == 5:
        labels = time_to_channels(labels)
      shape = common_layers.shape_list(labels)
      x = tf.expand_dims(labels, axis=-1)
      x = self.embed(x)
      target_codes = x
      print(x)
      if shape[2] == 1:
        self.is1d = True
      # Run encoder.
      x, encoder_layers = self.encoder(x)
      # Bottleneck.
      b, b_loss = self.bottleneck(x)
      xb_loss = 0.0
      b_shape = common_layers.shape_list(b)
      self._cur_bottleneck_tensor = b
      res_size = common_layers.shape_list(x)[-1]
      b = self.unbottleneck(b, res_size)
      if not is_training:
        x = b
      else:
        l = 2**hparams.num_hidden_layers
        warm_step = int(hparams.bottleneck_warmup_steps * 0.25 * l)
        nomix_p = common_layers.inverse_lin_decay(warm_step) + 0.01
        if common_layers.should_generate_summaries():
          tf.summary.scalar("nomix_p_bottleneck", nomix_p)
        rand = tf.random_uniform(common_layers.shape_list(x))
        # This is the distance between b and x. Having this as loss helps learn
        # the bottleneck function, but if we back-propagated to x it would be
        # minimized by just setting x=0 and b=0 -- so we don't want too much
        # of the influence of this, and we stop-gradient to not zero-out x.
        x_stop = tf.stop_gradient(x)
        xb_loss = tf.reduce_mean(tf.reduce_sum(
            tf.squared_difference(x_stop, b), axis=-1))
        # To prevent this loss from exploding we clip at 1, but anneal clipping.
        clip_max = 1.0 / common_layers.inverse_exp_decay(
            warm_step, min_value=0.001)
        xb_clip = tf.maximum(tf.stop_gradient(xb_loss), clip_max)
        xb_loss *= clip_max / xb_clip
        x = tf.where(tf.less(rand, nomix_p), b, x)
    else:
      if self._cur_bottleneck_tensor is None:
        b = self.sample()
      else:
        b = self._cur_bottleneck_tensor
      self._cur_bottleneck_tensor = b
      res_size = self.hparams.hidden_size * 2**self.hparams.num_hidden_layers
      res_size = min(res_size, hparams.max_hidden_size)
      x = self.unbottleneck(b, res_size)
    # Run decoder.
    x = self.decoder(x, encoder_layers)

    # Cut to the right size and mix before returning.
    res = x
    if hparams.mode != tf.estimator.ModeKeys.PREDICT:
      res = x[:, :shape[1], :shape[2], :]

    # Final dense layer.
    res = tf.layers.dense(
        res, self.num_channels * hparams.hidden_size, name="res_dense")

    output_shape = common_layers.shape_list(res)[:-1] + [
        self.num_channels, self.hparams.hidden_size
    ]
    res = tf.reshape(res, output_shape)

    if hparams.mode == tf.estimator.ModeKeys.PREDICT:
      reconstr = tf.layers.dense(res, self.num_channels, name="autoencoder_final")
      return reconstr, {"bottleneck_loss": 0.0}

    # Losses.
    losses = {
        "bottleneck_extra": b_loss,
        "bottleneck_l2": hparams.bottleneck_l2_factor * xb_loss
    }

    reconstr = tf.layers.dense(res, self.num_channels, name="autoencoder_final")
    reconstr = tf.reshape(reconstr, labels_shape)

    targets_loss = self.reconstruction_loss(reconstr, labels)
    losses["training"] = targets_loss

    self.image_summary("inputs", labels)
    self.image_summary("ae", reconstr)
    return reconstr, losses


@registry.register_model
class ContinuousAutoencoderResidual(ContinuousAutoencoderBasic):
  """Residual autoencoder."""

  def dropout(self, x):
    is_training = self.hparams.mode == tf.estimator.ModeKeys.TRAIN
    hparams = self.hparams
    if hparams.dropout <= 0.0 or not is_training:
      return x
    warm_step = hparams.bottleneck_warmup_steps * 2**hparams.num_hidden_layers
    dropout = common_layers.inverse_lin_decay(warm_step // 2) * hparams.dropout
    return common_layers.dropout_with_broadcast_dims(
        x, 1.0 - dropout, broadcast_dims=[-1])

  def encoder(self, x):
    with tf.variable_scope("encoder"):
      hparams = self.hparams
      layers = []
      kernel, strides = self._get_kernel_and_strides()
      residual_kernel = (hparams.residual_kernel_height,
                         hparams.residual_kernel_width)
      residual_kernel1d = (hparams.residual_kernel_height, 1)
      residual_kernel = residual_kernel1d if self.is1d else residual_kernel
      residual_conv = tf.layers.conv2d
      if hparams.residual_use_separable_conv:
        residual_conv = tf.layers.separable_conv2d
      # Down-convolutions.
      for i in range(hparams.num_hidden_layers):
        with tf.variable_scope("layer_%d" % i):
          x = self.make_even_size(x)
          layers.append(x)
          x = self.dropout(x)
          filters = hparams.hidden_size * 2**(i + 1)
          filters = min(filters, hparams.max_hidden_size)
          x = common_attention.add_timing_signal_nd(x)
          x = tf.layers.conv2d(
              x,
              filters,
              kernel,
              strides=strides,
              padding="SAME",
              activation=common_layers.belu,
              name="strided")
          y = x
          y = tf.nn.dropout(y, 1.0 - hparams.residual_dropout)
          for r in range(hparams.num_residual_layers):
            residual_filters = filters
            if r < hparams.num_residual_layers - 1:
              residual_filters = int(
                  filters * hparams.residual_filter_multiplier)
            y = residual_conv(
                y,
                residual_filters,
                residual_kernel,
                padding="SAME",
                activation=common_layers.belu,
                name="residual_%d" % r)
          x += y
          x = common_layers.layer_norm(x, name="ln")
      return x, layers

  def decoder(self, x, encoder_layers=None):
    with tf.variable_scope("decoder"):
      hparams = self.hparams
      is_training = self.hparams.mode == tf.estimator.ModeKeys.TRAIN
      kernel, strides = self._get_kernel_and_strides()
      residual_kernel = (hparams.residual_kernel_height,
                         hparams.residual_kernel_width)
      residual_kernel1d = (hparams.residual_kernel_height, 1)
      residual_kernel = residual_kernel1d if self.is1d else residual_kernel
      residual_conv = tf.layers.conv2d
      if hparams.residual_use_separable_conv:
        residual_conv = tf.layers.separable_conv2d
      # Up-convolutions.
      for i in range(hparams.num_hidden_layers):
        j = hparams.num_hidden_layers - i - 1
        if is_training:
          nomix_p = common_layers.inverse_lin_decay(
              int(hparams.bottleneck_warmup_steps * 0.25 * 2**j)) + 0.01
          if common_layers.should_generate_summaries():
            tf.summary.scalar("nomix_p_%d" % j, nomix_p)
        filters = hparams.hidden_size * 2**j
        filters = min(filters, hparams.max_hidden_size)
        with tf.variable_scope("layer_%d" % i):
          j = hparams.num_hidden_layers - i - 1
          x = tf.layers.conv2d_transpose(
              x,
              filters,
              kernel,
              strides=strides,
              padding="SAME",
              activation=common_layers.belu,
              name="strided")
          y = x
          for r in range(hparams.num_residual_layers):
            residual_filters = filters
            if r < hparams.num_residual_layers - 1:
              residual_filters = int(
                  filters * hparams.residual_filter_multiplier)
            y = residual_conv(
                y,
                residual_filters,
                residual_kernel,
                padding="SAME",
                activation=common_layers.belu,
                name="residual_%d" % r)
          x += tf.nn.dropout(y, 1.0 - hparams.residual_dropout)
          x = common_layers.layer_norm(x, name="ln")
          x = common_attention.add_timing_signal_nd(x)
          if encoder_layers is not None:
            enc_x = encoder_layers[j]
            enc_shape = common_layers.shape_list(enc_x)
            x_mix = x[:enc_shape[0], :enc_shape[1], :enc_shape[2], :]
            if is_training:  # Mix at the beginning of training.
              rand = tf.random_uniform(common_layers.shape_list(x_mix))
              x_mix = tf.where(tf.less(rand, nomix_p), x_mix, enc_x)
            x = x_mix
      return x


@registry.register_model
class ContinuousAutoencoderResidualVAE(ContinuousAutoencoderResidual):
  """Residual VAE autoencoder."""

  def bottleneck(self, x):
    hparams = self.hparams
    z_size = hparams.bottleneck_bits
    x_shape = common_layers.shape_list(x)
    with tf.variable_scope("vae"):
      mu = tf.layers.dense(x, z_size, name="mu")
      if hparams.mode != tf.estimator.ModeKeys.TRAIN:
        return mu, 0.0  # No sampling or kl loss on eval.
      log_sigma = tf.layers.dense(x, z_size, name="log_sigma")
      epsilon = tf.random_normal(x_shape[:-1] + [z_size])
      z = mu + tf.exp(log_sigma / 2) * epsilon
      kl = 0.5 * tf.reduce_mean(
          tf.expm1(log_sigma) + tf.square(mu) - log_sigma, axis=-1)
      free_bits = z_size // 4
      kl_loss = tf.reduce_mean(tf.maximum(kl - free_bits, 0.0))
    return z, kl_loss * hparams.kl_beta

  def sample(self, features=None, shape=None):
    del features
    hparams = self.hparams
    div_x = 2**hparams.num_hidden_layers
    div_y = 1 if self.is1d else 2**hparams.num_hidden_layers
    size = [
        hparams.batch_size, hparams.sample_height // div_x,
        hparams.sample_width // div_y, hparams.bottleneck_bits
    ]
    size = size if shape is None else shape
    return tf.random_normal(size)


@registry.register_model
class ContinuousAutoencoderBasicDiscrete(ContinuousAutoencoderBasic):
  """Discrete autoencoder."""

  def bottleneck(self, x):
    hparams = self.hparams
    x = tf.tanh(tf.layers.dense(x, hparams.bottleneck_bits, name="bottleneck"))
    d = x + tf.stop_gradient(2.0 * tf.to_float(tf.less(0.0, x)) - 1.0 - x)
    if hparams.mode == tf.estimator.ModeKeys.TRAIN:
      noise = tf.random_uniform(common_layers.shape_list(x))
      noise = 2.0 * tf.to_float(tf.less(hparams.bottleneck_noise, noise)) - 1.0
      d *= noise
    x = common_layers.mix(d, x, hparams.discretize_warmup_steps,
                          hparams.mode == tf.estimator.ModeKeys.TRAIN)
    return x, 0.0

  def sample(self, features=None, shape=None):
    del features
    hp = self.hparams
    div_x = 2**hp.num_hidden_layers
    div_y = 1 if self.is1d else 2**hp.num_hidden_layers
    size = [
        hp.batch_size, hp.sample_height // div_x, hp.sample_width // div_y,
        hp.bottleneck_bits
    ]
    size = size if shape is None else shape
    rand = tf.random_uniform(size)
    return 2.0 * tf.to_float(tf.less(0.5, rand)) - 1.0


@registry.register_model
class ContinuousAutoencoderResidualDiscrete(ContinuousAutoencoderResidual):
  """Discrete residual autoencoder."""

  def variance_loss(self, b):
    part = tf.random_uniform(common_layers.shape_list(b))
    selection = tf.to_float(tf.less(part, tf.random_uniform([])))
    selection_size = tf.reduce_sum(selection)
    part_avg = tf.abs(tf.reduce_sum(b * selection)) / (selection_size + 1)
    return part_avg

  def bottleneck(self, x, bottleneck_bits=None):  # pylint: disable=arguments-differ
    if bottleneck_bits is not None:
      old_bottleneck_bits = self.hparams.bottleneck_bits
      self.hparams.bottleneck_bits = bottleneck_bits
    res, loss = discretization.parametrized_bottleneck(x, self.hparams)
    if bottleneck_bits is not None:
      self.hparams.bottleneck_bits = old_bottleneck_bits
    return res, loss

  def unbottleneck(self, x, res_size, reuse=None):
    with tf.variable_scope("unbottleneck", reuse=reuse):
      return discretization.parametrized_unbottleneck(x, res_size, self.hparams)

  def sample(self, features=None, shape=None):
    del features
    hp = self.hparams
    div_x = 2**hp.num_hidden_layers
    div_y = 1 if self.is1d else 2**hp.num_hidden_layers
    size = [
        hp.batch_size, hp.sample_height // div_x, hp.sample_width // div_y,
        hp.bottleneck_bits
    ]
    size = size if shape is None else shape
    rand = tf.random_uniform(size)
    res = 2.0 * tf.to_float(tf.less(0.5, rand)) - 1.0
    # If you want to set some first bits to a fixed value, do this:
    # fixed = tf.zeros_like(rand) - 1.0
    # nbits = 3
    # res = tf.concat([fixed[:, :, :, :nbits], res[:, :, :, nbits:]], axis=-1)
    return res


@registry.register_model
class ContinuousAutoencoderOrderedDiscrete(ContinuousAutoencoderResidualDiscrete):
  """Ordered discrete autoencoder."""

  def bottleneck(self, x):  # pylint: disable=arguments-differ
    hparams = self.hparams
    if hparams.unordered:
      return super(AutoencoderOrderedDiscrete, self).bottleneck(x)
    noise = hparams.bottleneck_noise
    hparams.bottleneck_noise = 0.0  # We'll add noise below.
    x, loss = discretization.parametrized_bottleneck(x, hparams)
    hparams.bottleneck_noise = noise
    if hparams.mode == tf.estimator.ModeKeys.TRAIN:
      # We want a number p such that p^bottleneck_bits = 1 - noise.
      # So log(p) * bottleneck_bits = log(noise)
      log_p = tf.log1p(-float(noise) / 2) / float(hparams.bottleneck_bits)
      # Probabilities of flipping are p, p^2, p^3, ..., p^bottleneck_bits.
      noise_mask = 1.0 - tf.exp(tf.cumsum(tf.zeros_like(x) + log_p, axis=-1))
      # Having the no-noise mask, we can make noise just uniformly at random.
      ordered_noise = tf.random_uniform(tf.shape(x))
      # We want our noise to be 1s at the start and random {-1, 1} bits later.
      ordered_noise = tf.to_float(tf.less(noise_mask, ordered_noise))
      # Now we flip the bits of x on the noisy positions (ordered and normal).
      x *= 2.0 * ordered_noise - 1
    return x, loss


@registry.register_hparams
def continuous_autoencoder_basic():
  """Basic autoencoder model."""
  hparams = common_hparams.basic_params1()
  hparams.optimizer = "adam"
  hparams.learning_rate_constant = 0.0002
  hparams.learning_rate_warmup_steps = 500
  hparams.learning_rate_schedule = "constant * linear_warmup"
  hparams.label_smoothing = 0.0
  hparams.batch_size = 128
  hparams.hidden_size = 64
  hparams.num_hidden_layers = 5
  hparams.initializer = "uniform_unit_scaling"
  hparams.initializer_gain = 1.0
  hparams.weight_decay = 0.0
  hparams.kernel_height = 4
  hparams.kernel_width = 4
  hparams.dropout = 0.05
  hparams.add_hparam("max_hidden_size", 1024)
  hparams.add_hparam("bottleneck_bits", 128)
  hparams.add_hparam("bottleneck_shared_bits", 0)
  hparams.add_hparam("bottleneck_shared_bits_start_warmup", 0)
  hparams.add_hparam("bottleneck_shared_bits_stop_warmup", 0)
  hparams.add_hparam("bottleneck_noise", 0.1)
  hparams.add_hparam("bottleneck_warmup_steps", 2000)
  hparams.add_hparam("sample_height", 32)
  hparams.add_hparam("sample_width", 32)
  hparams.add_hparam("num_sliced_vecs", 20000)
  hparams.add_hparam("sliced_do_tanh", int(True))
  hparams.add_hparam("code_loss_factor", 1.0)
  hparams.add_hparam("bottleneck_l2_factor", 0.05)
  hparams.add_hparam("reconstruction_loss_sigma", 0.03)
  return hparams

@registry.register_hparams
def continuous_autoencoder_residual():
  """Residual autoencoder model."""
  hparams = continuous_autoencoder_basic()
  hparams.optimizer = "Adafactor"
  hparams.clip_grad_norm = 1.0
  hparams.learning_rate_constant = 0.5
  hparams.learning_rate_warmup_steps = 500
  hparams.learning_rate_schedule = "constant * linear_warmup * rsqrt_decay"
  hparams.num_hidden_layers = 5
  hparams.hidden_size = 64
  hparams.max_hidden_size = 1024
  hparams.add_hparam("num_residual_layers", 2)
  hparams.add_hparam("residual_kernel_height", 3)
  hparams.add_hparam("residual_kernel_width", 3)
  hparams.add_hparam("residual_filter_multiplier", 2.0)
  hparams.add_hparam("residual_dropout", 0.2)
  hparams.add_hparam("residual_use_separable_conv", int(True))
  hparams.add_hparam("kl_beta", 1.0)
  return hparams

@registry.register_hparams
def continuous_autoencoder_basic_discrete():
  """Basic autoencoder model."""
  hparams = continuous_autoencoder_basic()
  hparams.num_hidden_layers = 5
  hparams.hidden_size = 64
  hparams.bottleneck_bits = 1024
  hparams.bottleneck_noise = 0.1
  hparams.add_hparam("discretize_warmup_steps", 16000)
  return hparams

@registry.register_hparams
def continuous_autoencoder_residual_discrete():
  """Residual discrete autoencoder model."""
  hparams = continuous_autoencoder_residual()
  hparams.bottleneck_bits = 1024
  hparams.bottleneck_noise = 0.05
  hparams.add_hparam("discretize_warmup_steps", 16000)
  hparams.add_hparam("bottleneck_kind", "tanh_discrete")
  hparams.add_hparam("isemhash_noise_dev", 0.5)
  hparams.add_hparam("isemhash_mix_prob", 0.5)
  hparams.add_hparam("isemhash_filter_size_multiplier", 2.0)
  hparams.add_hparam("vq_beta", 0.25)
  hparams.add_hparam("vq_decay", 0.999)
  hparams.add_hparam("vq_epsilon", 1e-5)
  return hparams

@registry.register_hparams
def continuous_autoencoder_residual_discrete_big():
  """Residual discrete autoencoder model, big version."""
  hparams = continuous_autoencoder_residual_discrete()
  hparams.hidden_size = 128
  hparams.max_hidden_size = 4096
  hparams.bottleneck_noise = 0.1
  hparams.residual_dropout = 0.4
  return hparams

@registry.register_hparams
def continuous_autoencoder_ordered_discrete():
  """Ordered discrete autoencoder model."""
  hparams = continuous_autoencoder_residual_discrete()
  hparams.bottleneck_noise = 0.05  # Use 0.8 for ordered.
  hparams.add_hparam("unordered", True)
  return hparams

@registry.register_hparams
def continuous_autoencoder_ordered_discrete_vq():
  """Ordered discrete autoencoder model with VQ bottleneck."""
  hparams = continuous_autoencoder_ordered_discrete()
  hparams.bottleneck_kind = "vq"
  hparams.bottleneck_bits = 16
  return hparams

@registry.register_hparams
def continuous_autoencoder_discrete_cifar():
  """Discrete autoencoder model for compressing cifar."""
  hparams = continuous_autoencoder_ordered_discrete()
  hparams.bottleneck_noise = 0.0
  hparams.bottleneck_bits = 90
  hparams.num_hidden_layers = 2
  hparams.hidden_size = 256
  hparams.num_residual_layers = 4
  hparams.batch_size = 32
  hparams.learning_rate_constant = 1.0
  return hparams



"""Imports for problem modules."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib
import six
from six.moves import range  # pylint: disable=redefined-builtin

MODULES = [
    "galaxy2galaxy.data_generators.hsc"
]
# Modules that depend on galsim, only include if available
try:
  import galsim
  MODULES += ["galaxy2galaxy.data_generators.cosmos"]
except:
  print("Could not import GalSim, excluding some data generators")

ALL_MODULES = list(MODULES)


def _is_import_err_msg(err_str, module):
  parts = module.split(".")
  suffixes = [".".join(parts[i:]) for i in range(len(parts))]
  return err_str in (
      ["No module named %s" % suffix for suffix in suffixes] +
      ["No module named '%s'" % suffix for suffix in suffixes])


def _handle_errors(errors):
  """Log out and possibly reraise errors during import."""
  if not errors:
    return
  log_all = True  # pylint: disable=unused-variable
  err_msg = "G2G: skipped importing {num_missing} data_generators modules."
  print(err_msg.format(num_missing=len(errors)))
  for module, err in errors:
    err_str = str(err)
    if not _is_import_err_msg(err_str, module):
      print("From module %s" % module)
      raise err
    if log_all:
      print("Did not import module: %s; Cause: %s" % (module, err_str))


def import_modules(modules):
  errors = []
  for module in modules:
    try:
      importlib.import_module(module)
    except ImportError as error:
      errors.append((module, error))
  _handle_errors(errors)



"""Scripts to generate HSC samples from the HSC archive."""
import argparse
from unagi import hsc
from unagi import task

import os
import shutil
import astropy.units as u
from astropy.table import Table

def build_hsc_sample(sql_file,
                     out_dir,
                     tmp_dir,
                     filters='i',
                     cutout_size=10.0, # in arcsec
                     data_release='pdr2',
                     rerun='pdr2_wide',
                     nproc=2):
    """
    This function runs an sql query to extract a catalog, then
    proceeds to download cutouts in requested bands for all
    catalog entries.
    """
    # Creates output directory if doesn't already exists
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    # Create archive instance and login
    archive = hsc.Hsc(dr=data_release, rerun=rerun)

    catalog_filename = os.path.join(out_dir, 'catalog.fits')
    if os.path.isfile(catalog_filename):
        print("Readding hsc catalog from %s"%catalog_filename)
        catalog = Table.read(catalog_filename)
    else:
        # Query the database
        catalog = archive.sql_query(sql_file,
                                from_file=True,
                                verbose=True)

        print("Saving hsc catalog to %s"%catalog_filename)
        catalog.write(catalog_filename)

    # Query corresponding postage stamps
    cutouts_filename = task.hsc_bulk_cutout(catalog,
                                            cutout_size=cutout_size* u.Unit('arcsec'),
                                            filters=filters,
                                            archive=archive,
                                            nproc=nproc,
                                            tmp_dir=tmp_dir,
                                            output_dir=out_dir)
    # Rename the cutout file
    shutil.move(cutouts_filename, os.path.join(out_dir, 'cutouts.hdf'))



import contextlib
import tensorflow as tf
import tensorflow_gan as tfgan

import galaxy2galaxy.layers.spectral_ops as ops

def usample(x):
  """Upsamples the input volume.
  Args:
    x: The 4D input tensor.
  Returns:
    An upsampled version of the input tensor.
  """
  # Allow the batch dimension to be unknown at graph build time.
  _, image_height, image_width, n_channels = x.shape.as_list()
  # Add extra degenerate dimension after the dimensions corresponding to the
  # rows and columns.
  expanded_x = tf.expand_dims(tf.expand_dims(x, axis=2), axis=4)
  # Duplicate data in the expanded dimensions.
  after_tile = tf.tile(expanded_x, [1, 1, 2, 1, 2, 1])
  return tf.reshape(after_tile,
                    [-1, image_height * 2, image_width * 2, n_channels])

def up_block(x, out_channels, name, training=True):
  """Builds the residual blocks used in the generator.
  Args:
    x: The 4D input tensor.
    out_channels: Integer number of features in the output layer.
    name: The variable scope name for the block.
    training: Whether this block is for training or not.
  Returns:
    A `Tensor` representing the output of the operation.
  """
  with tf.variable_scope(name):
    bn0 = ops.BatchNorm(name='bn_0')
    bn1 = ops.BatchNorm(name='bn_1')
    x_0 = x
    x = tf.nn.relu(bn0(x))
    x = usample(x)
    x = ops.snconv2d(x, out_channels, 3, 3, 1, 1, training, 'snconv1')
    x = tf.nn.relu(bn1(x))
    x = ops.snconv2d(x, out_channels, 3, 3, 1, 1, training, 'snconv2')

    x_0 = usample(x_0)
    x_0 = ops.snconv2d(x_0, out_channels, 1, 1, 1, 1, training, 'snconv3')

    return x_0 + x

def dsample(x):
  """Downsamples the input volume by means of average pooling.
  Args:
    x: The 4D input tensor.
  Returns:
    An downsampled version of the input tensor.
  """
  xd = tf.nn.avg_pool(x, [1, 2, 2, 1], [1, 2, 2, 1], 'VALID')
  return xd

def down_block(x, out_channels, name, downsample=True, act=tf.nn.relu):
  """Builds the residual blocks used in the discriminator.
  Args:
    x: The 4D input vector.
    out_channels: Number of features in the output layer.
    name: The variable scope name for the block.
    downsample: If True, downsample the spatial size the input tensor by
                a factor of 2 on each side. If False, the spatial size of the
                input tensor is unchanged.
    act: The activation function used in the block.
  Returns:
    A `Tensor` representing the output of the operation.
  """
  with tf.variable_scope(name):
    input_channels = x.shape.as_list()[-1]
    x_0 = x
    x = act(x)
    x = ops.snconv2d(x, out_channels, 3, 3, 1, 1, name='sn_conv1')
    x = act(x)
    x = ops.snconv2d(x, out_channels, 3, 3, 1, 1, name='sn_conv2')
    if downsample:
      x = dsample(x)
    if downsample or input_channels != out_channels:
      x_0 = ops.snconv2d(x_0, out_channels, 1, 1, 1, 1, name='sn_conv3')
      if downsample:
        x_0 = dsample(x_0)
    return x_0 + x


def down_optimized_block(x, out_channels, name, act=tf.nn.relu):
  """Builds optimized residual blocks for downsampling.
  Compared with block, optimized_block always downsamples the spatial resolution
  by a factor of 2 on each side.
  Args:
    x: The 4D input vector.
    out_channels: Number of features in the output layer.
    name: The variable scope name for the block.
    act: The activation function used in the block.
  Returns:
    A `Tensor` representing the output of the operation.
  """
  with tf.variable_scope(name):
    x_0 = x
    x = ops.snconv2d(x, out_channels, 3, 3, 1, 1, name='sn_conv1')
    x = act(x)
    x = ops.snconv2d(x, out_channels, 3, 3, 1, 1, name='sn_conv2')
    x = dsample(x)
    x_0 = dsample(x_0)
    x_0 = ops.snconv2d(x_0, out_channels, 1, 1, 1, 1, name='sn_conv3')
    return x + x_0




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import tensorflow_gan as tfgan
from tensorflow.python.summary import summary

from tensor2tensor.utils import hparams_lib
from tensor2tensor.layers import common_layers
from tensor2tensor.layers import common_hparams

from galaxy2galaxy.utils import registry
from galaxy2galaxy.models.gan_utils import AbstractGAN
from galaxy2galaxy.models.gan_utils import SummaryType

from galaxy2galaxy.layers import spectral_ops as ops
from galaxy2galaxy.layers.common_layers import up_block, down_block, down_optimized_block


@registry.register_model
class SelfAttentionGan(AbstractGAN):
  """ Implementation of Self Attention GAN
  """

  def sample_noise(self):
    p = self.hparams
    shape = [p.batch_size, p.bottleneck_bits]
    z = tf.random.normal(shape, name='z0', dtype=tf.float32)
    return z

  def generator(self, code, mode):
    """Builds the generator segment of the graph, going from z -> G(z).
    Args:
    zs: Tensor representing the latent variables.
    gf_dim: The gf dimension.
    training: Whether in train mode or not. This affects things like batch
      normalization and spectral normalization.
    Returns:
    - The output layer of the generator.
    """
    training = (mode == tf.estimator.ModeKeys.TRAIN)
    p = self.hparams
    gf_dim = p.hidden_size
    with tf.variable_scope('generator', reuse=tf.AUTO_REUSE) as gen_scope:
      act0 = ops.snlinear(code, gf_dim * 16 * 4 * 4, training=training, name='g_snh0')
      act0 = tf.reshape(act0, [-1, 4, 4, gf_dim * 16])

      # pylint: disable=line-too-long
      act1 = up_block(act0, gf_dim * 16, 'g_block1', training)  # 8
      act2 = up_block(act1, gf_dim * 8, 'g_block2', training)  # 16
      act3 = up_block(act2, gf_dim * 4, 'g_block3', training)  # 32
      act3 = ops.sn_non_local_block_sim(act3, training, name='g_ops')  # 32
      act4 = up_block(act3, gf_dim * 2, 'g_block4', training)  # 64
      if p.noise_sigma >0:
        act4 = tf.concat([act4, p.noise_sigma*tf.random_normal((p.batch_size, 64, 64,3))],axis=-1)
      act5 = up_block(act4, gf_dim, 'g_block5', training)  # 128
      bn = ops.BatchNorm(name='g_bn')

      act5 = tf.nn.relu(bn(act5))
      act6 = ops.snconv2d(act5, 3, 3, 3, 1, 1, training, 'g_snconv_last')
      out = tf.nn.tanh(act6)
      return out

  def discriminator(self, image, conditioning, mode):
    """Builds the discriminator graph.
    Args:
        image: The current batch of images to classify as fake or real.
        df_dim: The df dimension.
        act: The activation function used in the discriminator.
      Returns:
        - A `Tensor` representing the logits of the discriminator.
        - A list containing all trainable varaibles defined by the model.
    """
    p = self.hparams
    df_dim = p.hidden_size
    training = (mode == tf.estimator.ModeKeys.TRAIN)
    act=tf.nn.relu
    with tf.variable_scope('discriminator', reuse=tf.AUTO_REUSE) as dis_scope:
      h0 = down_optimized_block(image, df_dim, 'd_optimized_block1', act=act)  # 64 * 64
      h1 = down_block(h0, df_dim * 2, 'd_block2', act=act)  # 32 * 32
      h1 = ops.sn_non_local_block_sim(h1, name='d_ops')  # 32 * 32
      h2 = down_block(h1, df_dim * 4, 'd_block3', act=act)  # 16 * 16
      h3 = down_block(h2, df_dim * 8, 'd_block4', act=act)  # 8 * 8
      h4 = down_block(h3, df_dim * 16, 'd_block5', act=act)  # 4 * 4
      h5 = down_block(h4, df_dim * 16, 'd_block6', downsample=False, act=act)
      h5_act = act(h5)
      h6 = tf.reduce_sum(h5_act, [1, 2])
      output = ops.snlinear(h6, 1, name='d_sn_linear')
    return output

  def generator_loss(self,*args, **kwargs):
    return tfgan.losses.wasserstein_hinge_generator_loss(*args, **kwargs)

  def discriminator_loss(selg, *args, **kwargs):
    return tfgan.losses.wasserstein_hinge_discriminator_loss(*args, **kwargs)

@registry.register_hparams
def sagan():
  """Basic parameters for 128x128 SAGAN."""
  hparams = common_hparams.basic_params1()
  hparams.optimizer = "adam"
  hparams.learning_rate = 0.0001
  hparams.learning_rate_constant = 0.0002
  hparams.learning_rate_warmup_steps = 500
  hparams.learning_rate_schedule = "constant * linear_warmup"
  hparams.label_smoothing = 0.0
  hparams.batch_size = 64
  hparams.hidden_size = 32
  hparams.initializer = "uniform_unit_scaling"
  hparams.initializer_gain = 1.0
  hparams.weight_decay = 1e-6
  hparams.kernel_height = 4
  hparams.kernel_width = 4
  hparams.add_hparam("bottleneck_bits", 128)
  hparams.add_hparam("generator_lr", 0.0001)
  hparams.add_hparam("discriminator_lr", 0.0004)
  hparams.add_hparam("beta1", 0.)
  hparams.add_hparam("noise_sigma", 0.)
  hparams.add_hparam("gen_steps", 1)
  hparams.add_hparam("disc_steps", 1)
  return hparams

@registry.register_hparams
def sagan_noise():
  """Basic parameters for 128x128 SAGAN."""
  hparams = common_hparams.basic_params1()
  hparams.optimizer = "adam"
  hparams.learning_rate = 0.0001
  hparams.learning_rate_constant = 0.0002
  hparams.learning_rate_warmup_steps = 500
  hparams.learning_rate_schedule = "constant * linear_warmup"
  hparams.label_smoothing = 0.0
  hparams.batch_size = 64
  hparams.hidden_size = 32
  hparams.initializer = "uniform_unit_scaling"
  hparams.initializer_gain = 1.0
  hparams.weight_decay = 1e-6
  hparams.kernel_height = 4
  hparams.kernel_width = 4
  hparams.add_hparam("bottleneck_bits", 128)
  hparams.add_hparam("generator_lr", 0.0001)
  hparams.add_hparam("discriminator_lr", 0.0004)
  hparams.add_hparam("beta1", 0.)
  hparams.add_hparam("noise_sigma", 0.1)
  hparams.add_hparam("gen_steps", 1)
  hparams.add_hparam("disc_steps", 3)
  return hparams



"""Access G2G Problems."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from galaxy2galaxy.data_generators import all_problems
from galaxy2galaxy.utils import registry

def problem(name):
  return registry.problem(name)

def available():
  return sorted(registry.list_problems())

# Import problem modules
_modules = list(all_problems.MODULES)

all_problems.import_modules(_modules)



"""Models defined in G2G. Imports here force registration."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from galaxy2galaxy.models import autoencoders
from galaxy2galaxy.models import gans



""" Utilities for astronomical images problems."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensor2tensor.data_generators import generator_utils
from tensor2tensor.data_generators import image_utils
from tensor2tensor.data_generators import problem
from tensor2tensor.layers import modalities
from tensor2tensor.utils import metrics

import tensorflow as tf



from tensorflow.contrib.slim.python.slim.data import data_decoder
from tensorflow.contrib.slim.python.slim.data.tfexample_decoder import ItemHandler
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import sparse_tensor
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import map_fn
from tensorflow.python.ops import image_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import parsing_ops
from tensorflow.python.ops import sparse_ops

class Image(ItemHandler):
  """ Item Handler copied from
  https://github.com/tensorflow/tensorflow/blob/master/tensorflow/contrib/slim/python/slim/data/tfexample_decoder.py
  There is a bug in tf 1.14 that prevents reading raw images with arbitry numbers
  of channels. This is has a fix for that.
  TODO: Switch back to upstream data handler code at next release
  """

  def __init__(self,
               image_key=None,
               format_key=None,
               shape=None,
               channels=3,
               dtype=dtypes.uint8,
               repeated=False,
               dct_method=''):
    """Initializes the image.
    Args:
      image_key: the name of the TF-Example feature in which the encoded image
        is stored.
      format_key: the name of the TF-Example feature in which the image format
        is stored.
      shape: the output shape of the image as 1-D `Tensor`
        [height, width, channels]. The image is reshaped
        accordingly.
      channels: the number of channels in the image.
      dtype: images will be decoded at this bit depth. Different formats
        support different bit depths.
          See tf.image.decode_image,
              tf.io.decode_raw,
      repeated: if False, decodes a single image. If True, decodes a
        variable number of image strings from a 1D tensor of strings.
      dct_method: An optional string. Defaults to empty string. It only takes
        effect when image format is jpeg, used to specify a hint about the
        algorithm used for jpeg decompression. Currently valid values
        are ['INTEGER_FAST', 'INTEGER_ACCURATE']. The hint may be ignored, for
        example, the jpeg library does not have that specific option.
    """
    if not image_key:
      image_key = 'image/encoded'
    if not format_key:
      format_key = 'image/format'

    super(Image, self).__init__([image_key, format_key])
    self._image_key = image_key
    self._format_key = format_key
    self._shape = shape
    self._channels = channels
    self._dtype = dtype
    self._repeated = repeated
    self._dct_method = dct_method

  def tensors_to_item(self, keys_to_tensors):
    """See base class."""
    image_buffer = keys_to_tensors[self._image_key]
    image_format = keys_to_tensors[self._format_key]

    if self._repeated:
      return map_fn.map_fn(lambda x: self._decode(x, image_format),
                           image_buffer, dtype=self._dtype)
    else:
      return self._decode(image_buffer, image_format)

  def _decode(self, image_buffer, image_format):
    """Decodes the image buffer.
    Args:
      image_buffer: The tensor representing the encoded image tensor.
      image_format: The image format for the image in `image_buffer`. If image
        format is `raw`, all images are expected to be in this format, otherwise
        this op can decode a mix of `jpg` and `png` formats.
    Returns:
      A tensor that represents decoded image of self._shape, or
      (?, ?, self._channels) if self._shape is not specified.
    """
    # TODO: Assert that the image format is raw
    image = parsing_ops.decode_raw(image_buffer, out_type=self._dtype)
    image = array_ops.reshape(image, self._shape)
    return image

class AstroImageProblem(problem.Problem):
  """Base class for image problems on astronomical images.
  """

  # START: Subclass interface
  def hparams(self, defaults, model_hparams):
    p = defaults
    p.img_len = 64

  @property
  def dataset_splits(self):
    """Splits of data to produce and number of output shards for each.
       Note that each shard will be produced in parallel if parallel generation
       is used
    """
    return [{
                "split": problem.DatasetSplit.TRAIN,
                "shards": 8,
            }, {
                "split": problem.DatasetSplit.EVAL,
                "shards": 2,
            }]


  @property
  def is_generate_per_split(self):
    """A single call to `generate_samples` generates for all `dataset_splits`.

    Set to True if you already have distinct subsets of data for each dataset
    split specified in `self.dataset_splits`. `self.generate_samples` will be
    called once for each split.

    Set to False if you have a unified dataset that you'd like to have split out
    into training and evaluation data automatically. `self.generate_samples`
    will be called only once and the data will be sharded across the dataset
    splits specified in `self.dataset_splits`.

    Returns:
      bool
    """
    raise NotImplementedError

  @property
  def num_bands(self):
    """Number of bands."""
    raise NotImplementedError

  def generator(self, data_dir, tmp_dir, dataset_split, task_id=-1):
    """Generate examples of images.

    Each yielded dict will be made into a single example.

    This method is typically called once per split in `self.dataset_splits`
    unless `self.is_generate_per_split=False`.

    Args:
      data_dir: final data directory.
      tmp_dir: temporary directory that you can use for downloading and scratch.
      dataset_split: problem.DatasetSplit, which data split to generate samples
        for (for example, training and evaluation).
      task_id: number of the task, if parallel execution.

    Yields:
      dict
    """
    raise NotImplementedError

  def prepare_to_generate(self, data_dir, tmp_dir):
    """Prepare to generate data in parallel on different processes.

    This function is called if multiprocess_generate is True.

    Some things that might need to be done once are downloading the data
    if it is not yet downloaded, and building the vocabulary.

    Args:
      data_dir: a string
      tmp_dir: a string
    """
    pass

  @property
  def multiprocess_generate(self):
    """Whether to generate the data in multiple parallel processes."""
    return False

  @property
  def already_shuffled(self):
    return False
  # END: Subclass interface

  @property
  def num_train_shards(self):
    return self.dataset_splits[0]["shards"]

  @property
  def num_dev_shards(self):
    return self.dataset_splits[1]["shards"]

  @property
  def num_generate_tasks(self):
    """Needed if multiprocess_generate is True."""
    return self.num_train_shards + self.num_dev_shards

  def generate_data(self, data_dir, tmp_dir, task_id=-1):
    """
    Generates training/dev data.

    Args:
      data_dir: a string
      tmp_dir: a string
      task_id: an optional integer
    Returns:
      shard or shards for which data was generated.
    """

    # In case of parallel execution, each shard is generated by a different
    # process
    if self.multiprocess_generate:
      tf.logging.info("generate_data task_id=%s" % task_id)
      assert task_id >= 0 and task_id < self.num_generate_tasks
      if task_id < self.num_train_shards:
        out_file = self.training_filepaths(
            data_dir, self.num_train_shards, shuffled=False)[task_id]
        dataset_split = problem.DatasetSplit.TRAIN
      else:
        out_file = self.dev_filepaths(
            data_dir, self.num_dev_shards,
            shuffled=False)[task_id - self.num_train_shards]
        dataset_split = problem.DatasetSplit.EVAL
      generator_utils.generate_files(
          self.generator(data_dir, tmp_dir, dataset_split, task_id), [out_file])
      generator_utils.shuffle_dataset([out_file])
    else:
      filepath_fns = {
            problem.DatasetSplit.TRAIN: self.training_filepaths,
            problem.DatasetSplit.EVAL: self.dev_filepaths,
            problem.DatasetSplit.TEST: self.test_filepaths,
            }

      split_paths = [(split["split"], filepath_fns[split["split"]](
            data_dir, split["shards"], shuffled=self.already_shuffled))
                       for split in self.dataset_splits]

      all_paths = []
      for _, paths in split_paths:
        all_paths.extend(paths)

      if self.is_generate_per_split:
        for split, paths in split_paths:
          generator_utils.generate_files(
              self.generator(data_dir, tmp_dir, split), paths)
      else:
        generator_utils.generate_files(
            self.generator(
                data_dir, tmp_dir, problem.DatasetSplit.TRAIN), all_paths)

      generator_utils.shuffle_dataset(all_paths)

  def example_reading_spec(self):
    """Define how data is serialized to file and read back.

    Returns:
      data_fields: A dictionary mapping data names to its feature type.
      data_items_to_decoders: A dictionary mapping data names to TF Example
         decoders, to be used when reading back TF examples from disk.
    """
    p = self.get_hparams()

    data_fields = {
        "image/encoded": tf.FixedLenFeature((), tf.string),
        "image/format": tf.FixedLenFeature((), tf.string),
    }

    # Adds additional attributes to be decoded as specified in the configuration
    if hasattr(p, 'attributes'):
        for k in p.attributes:
            data_fields['attrs/'+k] = tf.FixedLenFeature([], tf.float32, -1)

    data_items_to_decoders = {
        "inputs": Image( # TODO: switch back to tf.contrib.slim.tfexample_decoder.
                image_key="image/encoded",
                format_key="image/format",
                channels=self.num_bands,
                shape=[p.img_len, p.img_len, self.num_bands],
                dtype=tf.float32),
    }

    if hasattr(p, 'attributes'):
        for k in p.attributes:
            data_items_to_decoders[k] = tf.contrib.slim.tfexample_decoder.Tensor('attrs/'+k)

    return data_fields, data_items_to_decoders

  def eval_metrics(self):
    eval_metrics = []
    return eval_metrics

  @property
  def decode_hooks(self):
    return [image_utils.convert_predictions_to_image_summaries]



""" HSC Datasets """

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import hsc_utils
from . import astroimage_utils

from tensor2tensor.data_generators import generator_utils
from tensor2tensor.data_generators import image_utils
from tensor2tensor.data_generators import problem
from tensor2tensor.layers import modalities, common_layers
from tensor2tensor.utils import metrics

from galaxy2galaxy.utils import registry

from scipy.ndimage import gaussian_filter

import tensorflow as tf
import numpy as np
import fits2hdf.pyhdfits as fits
from astropy.table import Table
from astropy.visualization import make_lupton_rgb
import h5py
import os

# HSC default pixel scale
_HSC_PIXEL_SCALE=0.168 #arcsec
# Path to sql files for HSC samples
_HSC_SAMPLE_SQL_DIR=os.path.join(os.path.dirname(os.path.realpath(__file__)), 'hsc_utils/queries')

def _resize_image(im, size):
  centh = im.shape[0]/2
  centw = im.shape[1]/2
  lh, rh = int(centh-size/2), int(centh+size/2)
  lw, rw = int(centw-size/2), int(centw+size/2)
  cropped = im[lh:rh, lw:rw, :]
  assert cropped.shape[0]==size and cropped.shape[1]==size, f"Wrong size! Still {cropped.shape}"
  return cropped

@registry.register_problem
class Img2imgHSC(astroimage_utils.AstroImageProblem):
  """Base class for image problems created from HSC Public Data Release.
  """

  # START: Subclass interface
  def hparams(self, defaults, model_hparams):
    p = defaults
    p.img_len = 64
    p.filters = ['HSC-G', 'HSC-R', 'HSC-I', 'HSC-Z']
    p.sql_file = os.path.join(_HSC_SAMPLE_SQL_DIR, 'hsc_pdr2_wide_img2img.sql')
    p.data_release = 'pdr2'
    p.rerun = 'pdr2_wide'

  @property
  def num_bands(self):
    """Number of bands."""
    p = self.get_hparams()
    return len(p.filters)

  def generator(self, data_dir, tmp_dir, dataset_split, task_id=-1):
    """ Generator yielding individual postage stamps.
    """
    p = self.get_hparams()

    # Step 1: Maybe regenerate the dataset
    self.maybe_build_dataset(tmp_dir)
    import timeit

    # Step 2: Extract postage stamps, resize them to requested size
    catalog = Table.read(os.path.join(tmp_dir, 'catalog.fits'))
    with h5py.File(os.path.join(tmp_dir, 'cutouts.hdf'),'r') as cutouts:
        # Loop through the examples, resize cutout to desired size
        for row in catalog:
          cutout = cutouts[str(row['object_id'])]
          im = [cutout[f]['HDU0']['DATA'][:] for f in p.filters]

          try:
              im = np.stack(im, axis=-1).astype('float32')
          except:
              print('Failure to stack bands', [i.shape for i in im])
              continue

          # Images may not have exactly the right number of pixels
          im = _resize_image(im, p.img_len)

          serialized_output = {"image/encoded": [im.tostring()],
                               "image/format": ["raw"]}

          # If attributes are requested, let's add them to the dataset
          if hasattr(p, 'attributes'):
            for k in p.attributes:
              serialized_output['attrs/'+k] = [np.asscalar(row[k])]

          yield serialized_output

  def preprocess_example(self, example, unused_mode, unused_hparams):
    """ Preprocess the examples, can be used for further augmentation or
    image standardization.
    """
    p = self.get_hparams()
    image = example["inputs"]

    # TODO: apply some pre-processing/normalization to the images

    if hasattr(p, 'attributes'):
      example["attributes"] = tf.stack([example[k] for k in p.attributes])

    example["inputs"] = image
    example["targets"] = image
    return example

  # END: Subclass interface
  @property
  def is_generate_per_split(self):
    return False

  def maybe_build_dataset(self, tmp_dir):
    """ Check that the HSC dataset is downloaded and ready in tmp_dir, otherwise
    redownload it from the server.
    """
    p = self.get_hparams()
    if (not os.path.isfile(os.path.join(tmp_dir, 'catalog.fits')) or
        not os.path.isfile(os.path.join(tmp_dir, 'cutouts.hdf')) ):
      hsc_utils.build_hsc_sample(p.sql_file,
                                 out_dir=tmp_dir,
                                 tmp_dir=os.path.join(tmp_dir,'tmp'),
                                 cutout_size=(1.5+p.img_len)*_HSC_PIXEL_SCALE/2, # Requesting slightly larger pixel size to avoid roundoff errors
                                 filters=p.filters,
                                 data_release=p.data_release,
                                 rerun=p.rerun)

@registry.register_problem
class Img2imgHSCAnomaly(Img2imgHSC):
  """ Dataset for anomaly detection on HSC data.
  """

  def hparams(self, defaults, model_hparams):
    p = defaults
    p.img_len = 128
    p.filters = ['HSC-G', 'HSC-R', 'HSC-I']
    p.sql_file = os.path.join(_HSC_SAMPLE_SQL_DIR, 'hsc_pdr2_wide_anomaly.sql')
    p.data_release = 'pdr2'
    p.rerun = 'pdr2_wide'
    p.attributes = ['g_cmodel_mag', 'r_cmodel_mag', 'i_cmodel_mag', 'z_cmodel_mag']
    p.modality = {"inputs": modalities.ModalityType.IDENTITY,
                  "attributes":  modalities.ModalityType.IDENTITY,
                  "targets": modalities.ModalityType.IDENTITY}
    p.vocab_size = {"inputs": None,
                    "attributes": None,
                    "targets": None}

  def preprocess_example(self, example, unused_mode, unused_hparams):
    """ Luptonize the examples, so that we can use t2t models easily
    """
    p = self.get_hparams()
    image = example["inputs"]

    # Apply Luptonic Asinh stretch, and return uint8 rgb images
    def my_func(x):
      return make_lupton_rgb(x[...,2], x[...,1], x[...,0], Q=15, stretch=0.5, minimum=0)

    int_image = tf.py_func(my_func, [image], tf.uint8)
    int_image.set_shape(image.shape)
    image = common_layers.convert_rgb_to_symmetric_real(int_image)

    if hasattr(p, 'attributes'):
      example["attributes"] = tf.stack([example[k] for k in p.attributes])

    example["inputs"] = image
    example["targets"] = image
    return example


@registry.register_problem
class Img2photozHSC(Img2imgHSC):
  """ Dataset for photoz estimation on HSC data.
  """

  def hparams(self, defaults, model_hparams):
    p = defaults
    p.img_len = 64
    p.filters = ['HSC-G', 'HSC-R', 'HSC-I', 'HSC-Z', 'HSC-Y']
    p.sql_file = os.path.join(_HSC_SAMPLE_SQL_DIR, 'hsc_pdr2_wide_photoz.sql')
    p.data_release = 'pdr2'
    p.rerun = 'pdr2_wide'
    p.attributes = ['a_g', 'a_r', 'a_i', 'a_z', 'a_y',
                   'g_cmodel_mag', 'r_cmodel_mag', 'i_cmodel_mag', 'z_cmodel_mag',
                   'specz_redshift', 'specz_redshift_err']
    p.modality = {"inputs": modalities.ModalityType.IDENTITY,
                  "attributes":  modalities.ModalityType.IDENTITY,
                  "targets": modalities.ModalityType.IDENTITY}
    p.vocab_size = {"inputs": None,
                    "attributes": None,
                    "targets": None}

  def preprocess_example(self, example, unused_mode, unused_hparams):
    """ Luptonize the examples, so that we can use t2t models easily
    """
    p = self.get_hparams()
    image = example["inputs"]

    if hasattr(p, 'attributes'):
      example["attributes"] = tf.stack([example[k] for k in p.attributes])

    example["inputs"] = image
    example["targets"] = example['specz_redshift']
    return example



"""HST/ACS Cosmos generator."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import galsim_utils
from . import astroimage_utils

from tensor2tensor.data_generators import generator_utils
from tensor2tensor.data_generators import image_utils
from tensor2tensor.data_generators import problem
from tensor2tensor.layers import modalities
from tensor2tensor.utils import metrics

from galaxy2galaxy.utils import registry

import tensorflow as tf
import galsim


@registry.register_problem
class Img2imgCosmos(galsim_utils.GalsimProblem):
  """
  Img2img problem on GalSim's COSMOS 25.2 sample, at native pixel resolution,
  on 64px postage stamps.
  """

  @property
  def dataset_splits(self):
    """Splits of data to produce and number of output shards for each.
       Note that each shard will be produced in parallel.
       We are going to split the GalSim data into shards of 1000 galaxies each,
       with 80 shards for training, 2 shards for validation.
    """
    return [{
        "split": problem.DatasetSplit.TRAIN,
        "shards": 80,
    }, {
        "split": problem.DatasetSplit.EVAL,
        "shards": 2,
    }]

  def hparams(self, defaults, model_hparams):
    p = defaults
    p.pixel_scale = 0.03
    p.img_len = 64
    p.example_per_shard = 1000
    p.modality = {"inputs": modalities.ModalityType.IDENTITY,
                  "targets": modalities.ModalityType.IDENTITY}
    p.vocab_size = {"inputs": None,
                    "targets": None}

  @property
  def num_bands(self):
    """Number of bands."""
    return 1

  def generator(self, data_dir, tmp_dir, dataset_split, task_id=-1):
    """
    Generates and yields postage stamps obtained with GalSim.
    """
    p = self.get_hparams()
    try:
        # try to use default galsim path to the data
        catalog = galsim.COSMOSCatalog()
    except:
        # If that fails, tries to use the specified tmp_dir
        catalog = galsim.COSMOSCatalog(dir=tmp_dir+'/COSMOS_25.2_training_sample')

    # Create a list of galaxy indices for this task, remember, there is a task
    # per shard, each shard is 1000 galaxies.
    assert(task_id > -1)
    index = range(task_id*p.example_per_shard,
                  min((task_id+1)*p.example_per_shard, catalog.getNObjects()))

    # Extracts additional information about the galaxies
    cat_param = catalog.param_cat[catalog.orig_index]
    from numpy.lib.recfunctions import append_fields
    import numpy as np

    bparams = cat_param['bulgefit']
    sparams = cat_param['sersicfit']
    cat_param = append_fields(cat_param, 'bulge_q', bparams[:,11])
    cat_param = append_fields(cat_param, 'bulge_beta', bparams[:,15])
    cat_param = append_fields(cat_param, 'disk_q', bparams[:,3])
    cat_param = append_fields(cat_param, 'disk_beta', bparams[:,7])
    cat_param = append_fields(cat_param, 'bulge_hlr', cat_param['hlr'][:,1])
    cat_param = append_fields(cat_param, 'bulge_flux_log10', np.where(cat_param['use_bulgefit'] ==1, np.log10(cat_param['flux'][:,1]), np.zeros(len(cat_param) )))
    cat_param = append_fields(cat_param, 'disk_hlr', cat_param['hlr'][:,2])
    cat_param = append_fields(cat_param, 'disk_flux_log10', np.where(cat_param['use_bulgefit'] ==1, np.log10(cat_param['flux'][:,2]), np.log10(cat_param['flux'][:,0])))

    for ind in index:
      # Draw a galaxy using GalSim, any kind of operation can be done here
      gal = catalog.makeGalaxy(ind, noise_pad_size=p.img_len * p.pixel_scale)

      # We apply the orginal psf
      psf = gal.original_psf

      # We save the corresponding attributes for this galaxy
      if hasattr(p, 'attributes'):
        params = cat_param[ind]
        attributes = {k: params[k] for k in p.attributes}
      else:
        attributes = None

      # Utility function encodes the postage stamp for serialized features
      yield galsim_utils.draw_and_encode_stamp(gal, psf,
                                               stamp_size=p.img_len,
                                               pixel_scale=p.pixel_scale,
                                               attributes=attributes)

  def preprocess_example(self, example, unused_mode, unused_hparams):
    """ Preprocess the examples, can be used for further augmentation or
    image standardization.
    """
    p = self.get_hparams()
    image = example["inputs"]

    # Clip to 1 the values of the image
    image = tf.clip_by_value(image, -1, 1)

    # Aggregate the conditions
    if hasattr(p, 'attributes'):
      example['attributes'] = tf.stack([example[k] for k in p.attributes])

    example["inputs"] = image
    example["targets"] = image
    return example

@registry.register_problem
class Attrs2imgCosmos(Img2imgCosmos):
  """ Conditional image generation problem based on COSMOS sample.
  """

  def hparams(self, defaults, model_hparams):
    p = defaults
    p.pixel_scale = 0.03
    p.img_len = 64
    p.example_per_shard = 1000
    p.modality = {"inputs": modalities.ModalityType.IDENTITY,
                  "attributes":  modalities.ModalityType.IDENTITY,
                  "targets": modalities.ModalityType.IDENTITY}
    p.vocab_size = {"inputs": None,
                    "attributes": None,
                    "targets": None}
    p.attributes = ['mag_auto', 'flux_radius', 'zphot', 'bulge_q', 'bulge_beta' ,
                    'disk_q', 'disk_beta', 'bulge_hlr', 'disk_hlr']

@registry.register_problem
class Img2imgCosmos32(Img2imgCosmos):
  """ Smaller version of the Img2imgCosmos problem, at half the pixel
  resolution
  """

  def hparams(self, defaults, model_hparams):
    p = defaults
    p.pixel_scale = 0.06
    p.img_len = 32
    p.example_per_shard = 1000
    p.modality = {"inputs": modalities.ModalityType.IDENTITY,
                  "targets": modalities.ModalityType.IDENTITY}
    p.vocab_size = {"inputs": None,
                    "targets": None}

@registry.register_problem
class Attrs2imgCosmos32(Attrs2imgCosmos):
  """ Lower resolution equivalent of conditional generation problem.
  """

  def hparams(self, defaults, model_hparams):
    p = defaults
    p.pixel_scale = 0.06
    p.img_len = 32
    p.example_per_shard = 1000
    p.modality = {"inputs": modalities.ModalityType.IDENTITY,
                  "attributes":  modalities.ModalityType.IDENTITY,
                  "targets": modalities.ModalityType.IDENTITY}
    p.vocab_size = {"inputs": None,
                    "attributes": None,
                    "targets": None}
    p.attributes = ['mag_auto', 'flux_radius', 'zphot', 'bulge_q', 'bulge_beta' ,
                    'disk_q', 'disk_beta', 'bulge_hlr', 'disk_hlr']



# coding=utf-8
# Copyright 2019 The TensorFlow GAN Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Spectral normalization ops borrowed from TF-GAN until next release is available"""

import contextlib
import tensorflow as tf
import tensorflow_gan as tfgan


sn_gettr = tfgan.features.spectral_normalization_custom_getter


def snconv2d(input_, output_dim, k_h=3, k_w=3, d_h=2, d_w=2, training=True,
             name='snconv2d'):
  """Creates a 2d conv-layer with Spectral Norm applied to the weights.

  Args:
    input_: 4D input tensor (batch size, height, width, channel).
    output_dim: Number of features in the output layer.
    k_h: The height of the convolutional kernel.
    k_w: The width of the convolutional kernel.
    d_h: The height stride of the convolutional kernel.
    d_w: The width stride of the convolutional kernel.
    training: If `True`, add the spectral norm assign ops.
    name: The name of the variable scope.
  Returns:
    conv: The normalized tensor.

  """
  with tf.compat.v1.variable_scope(
      name, custom_getter=sn_gettr(training=training)):
    return tf.layers.conv2d(
        input_,
        filters=output_dim,
        kernel_size=(k_h, k_w),
        strides=(d_h, d_w),
        padding='same',
        activation=None,
        use_bias=True,
        kernel_initializer=tf.contrib.layers.xavier_initializer(),
        bias_initializer=tf.compat.v1.initializers.zeros(),
        name=name)


def snlinear(x, output_size, bias_start=0.0, training=True, name='snlinear'):
  """Creates a linear layer with Spectral Normalization applied.

  Args:
    x: 2D input tensor (batch size, features).
    output_size: Integer number of features in output of layer.
    bias_start: Float to which bias parameters are initialized.
    training: If `True`, add the spectral norm assign ops.
    name: Optional, variable scope to put the layer's parameters into.
  Returns:
    The normalized output tensor of the linear layer.
  """
  with tf.compat.v1.variable_scope(
      name, custom_getter=sn_gettr(training=training)):
    return tf.layers.dense(
        x,
        output_size,
        activation=None,
        use_bias=True,
        kernel_initializer=tf.contrib.layers.xavier_initializer(),
        bias_initializer=tf.compat.v1.initializers.constant(bias_start))


def sn_embedding(x, number_classes, embedding_size, training=True,
                 name='snembedding'):
  """Creates an embedding lookup with Spectral Normalization applied.

  Args:
    x: 1D input tensor (batch size, ).
    number_classes: The number of classes.
    embedding_size: The length of the embeddding vector for each class.
    training: If `True`, add the spectral norm assign ops.
    name: Optional, variable scope to put the layer's parameters into
  Returns:
    The output tensor (batch size, embedding_size).
  """
  with tf.compat.v1.variable_scope(name):
    embedding_map = tf.compat.v1.get_variable(
        name='embedding_map',
        shape=[number_classes, embedding_size],
        initializer=tf.contrib.layers.xavier_initializer())
    embedding_map_bar_transpose = tfgan.features.spectral_normalize(
        tf.transpose(a=embedding_map), training=training)
    embedding_map_bar = tf.transpose(a=embedding_map_bar_transpose)
    return tf.nn.embedding_lookup(params=embedding_map_bar, ids=x)


class ConditionalBatchNorm(object):
  """Conditional Batch Normalization.

  The same as normal Batch Normalization, but there is a different (gamma, beta)
  pair for each possible category.
  For each  class, it has a specific gamma and beta as normalization variable.
  """

  # TODO(augustusodena) Merge conditional batch norm with batch norm.
  # TODO(augustusodena) Use more sophisticated FilM layer here.
  # TODO(augustusodena) Why does this need to be a class?
  def __init__(self, num_categories, name='conditional_batch_norm'):
    """Inits the object.

    This is just a setter.

    Args:
      num_categories: Integer number of classes (and gamma, beta pairs).
      name: String name to be used for scoping.
    Returns:
      Initialized object.
    """
    with tf.compat.v1.variable_scope(name):
      self.name = name
      self.num_categories = num_categories

  def __call__(self, inputs, labels):
    """Adds Conditional Batch norm to the TF Graph.

    Args:
      inputs: Tensor of inputs (e.g. images).
      labels: Tensor of labels - same first dimension as inputs.
    Returns:
      Output tensor.
    """
    inputs = tf.convert_to_tensor(value=inputs)
    inputs_shape = inputs.get_shape()
    params_shape = inputs_shape[-1:]
    axis = [0, 1, 2]
    shape = tf.TensorShape([self.num_categories]).concatenate(params_shape)

    with tf.compat.v1.variable_scope(self.name):
      self.gamma = tf.compat.v1.get_variable(
          'gamma', shape, initializer=tf.compat.v1.initializers.ones())
      self.beta = tf.compat.v1.get_variable(
          'beta', shape, initializer=tf.compat.v1.initializers.zeros())
      beta = tf.gather(self.beta, labels)
      beta = tf.expand_dims(tf.expand_dims(beta, 1), 1)
      gamma = tf.gather(self.gamma, labels)
      gamma = tf.expand_dims(tf.expand_dims(gamma, 1), 1)

      mean, variance = tf.nn.moments(x=inputs, axes=axis, keepdims=True)
      outputs = tf.nn.batch_normalization(
          inputs, mean, variance, beta, gamma, variance_epsilon=1e-5)
      outputs.set_shape(inputs_shape)
      return outputs


class BatchNorm(object):
  """Batch Normalization.

  This is just vanilla batch normalization.
  """

  def __init__(self, name='batch_norm'):
    """Inits the object.

    This is just a setter.

    Args:
      name: String name to be used for scoping.
    Returns:
      Initialized object.
    """
    with tf.compat.v1.variable_scope(name):
      self.name = name

  def __call__(self, inputs):
    """Adds Batch Norm to the TF Graph.

    Args:
      inputs: Tensor of inputs (e.g. images).
    Returns:
      Output tensor.
    """
    inputs = tf.convert_to_tensor(value=inputs)
    inputs_shape = inputs.get_shape().as_list()
    params_shape = inputs_shape[-1]
    axis = [0, 1, 2]
    shape = tf.TensorShape([params_shape])
    with tf.compat.v1.variable_scope(self.name):
      self.gamma = tf.compat.v1.get_variable(
          'gamma', shape, initializer=tf.compat.v1.initializers.ones())
      self.beta = tf.compat.v1.get_variable(
          'beta', shape, initializer=tf.compat.v1.initializers.zeros())
      beta = self.beta
      gamma = self.gamma

      mean, variance = tf.nn.moments(x=inputs, axes=axis, keepdims=True)
      outputs = tf.nn.batch_normalization(
          inputs, mean, variance, beta, gamma, variance_epsilon=1e-5)
      outputs.set_shape(inputs_shape)
      return outputs


def sn_conv1x1(x, output_dim, training=True, name='sn_conv1x1'):
  """Builds graph for a spectrally normalized 1 by 1 convolution.

  This is used in the context of non-local networks to reduce channel count for
  strictly computational reasons.

  Args:
    x: A 4-D tensorflow tensor.
    output_dim: An integer representing desired channel count in the output.
    training: If `True`, add the spectral norm assign ops.
    name: String to pass to the variable scope context.
  Returns:
    A new volume with the same batch, height, and width as the input.
  """
  with tf.compat.v1.variable_scope(
      name, custom_getter=sn_gettr(training=training)):
    w = tf.compat.v1.get_variable(
        'weights', [1, 1, x.get_shape()[-1], output_dim],
        initializer=tf.contrib.layers.xavier_initializer())
    conv = tf.nn.conv2d(
        input=x, filters=w, strides=[1, 1, 1, 1], padding='SAME')
    return conv


def sn_non_local_block_sim(x, training=True, name='sn_nonlocal'):
  """Builds graph for the self-attention block.

  This is one third of the tricks from the SAGAN paper.

  Args:
    x: A 4-D tensorflow tensor.
    training: If `True`, add the spectral norm assign ops.
    name: String to pass to the variable scope context.
  Returns:
    A new volume with self-attention having been applied.
  """
  with tf.compat.v1.variable_scope(name):
    _, h, w, num_channels = x.shape.as_list()
    location_num = h * w
    downsampled_num = location_num // 4

    # theta path
    theta = sn_conv1x1(x, num_channels // 8, training, 'sn_conv_theta')
    theta = tf.reshape(
        theta, [-1, location_num, num_channels // 8])

    # phi path
    phi = sn_conv1x1(x, num_channels // 8, training, 'sn_conv_phi')
    phi = tf.layers.max_pooling2d(inputs=phi, pool_size=[2, 2], strides=2)
    phi = tf.reshape(
        phi, [-1, downsampled_num, num_channels // 8])

    attn = tf.matmul(theta, phi, transpose_b=True)
    attn = tf.nn.softmax(attn)

    # g path
    g = sn_conv1x1(x, num_channels // 2, training, 'sn_conv_g')
    g = tf.layers.max_pooling2d(inputs=g, pool_size=[2, 2], strides=2)
    g = tf.reshape(
        g, [-1, downsampled_num, num_channels // 2])

    attn_g = tf.matmul(attn, g)
    attn_g = tf.reshape(attn_g, [-1, h, w, num_channels // 2])
    sigma = tf.compat.v1.get_variable(
        'sigma_ratio', [], initializer=tf.compat.v1.initializers.constant(0.0))
    attn_g = sn_conv1x1(attn_g, num_channels, training, 'sn_conv_attn')
    return x + sigma * attn_g


@contextlib.contextmanager
def variables_on_gpu0():
  """Put variables on GPU."""
  old_fn = tf.compat.v1.get_variable

  def new_fn(*args, **kwargs):
    with tf.device('/gpu:0'):
      return old_fn(*args, **kwargs)

  tf.compat.v1.get_variable = new_fn
  yield
  tf.compat.v1.get_variable = old_fn


def avg_grads(tower_grads):
  """Calculate the average gradient for each shared variable across all towers.

  Note that this function provides a synchronization point across all towers.

  Args:
    tower_grads: List of lists of (gradient, variable) tuples. The outer list is
      over individual gradients. The inner list is over the gradient calculation
      for each tower.

  Returns:
     List of pairs of (gradient, variable) where the gradient has been averaged
     across all towers.
  """
  average_grads = []
  for grad_and_vars in zip(*tower_grads):
    # Note that each grad_and_vars looks like the following:
    #   ((grad0_gpu0, var0_gpu0), ... , (grad0_gpuN, var0_gpuN))
    grads = []
    for g, _ in grad_and_vars:
      # Add 0 dimension to the gradients to represent the tower.
      expanded_g = tf.expand_dims(g, 0)

      # Append on a 'tower' dimension which we will average over below.
      grads.append(expanded_g)

    # Average over the 'tower' dimension.
    grad = tf.concat(grads, 0)
    grad = tf.reduce_mean(input_tensor=grad, axis=0)

    # Keep in mind that the Variables are redundant because they are shared
    # across towers. So .. we will just return the first tower's pointer to
    # the Variable.
    v = grad_and_vars[0][1]
    grad_and_var = (grad, v)
    average_grads.append(grad_and_var)
  return average_grads



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensor2tensor.utils import registry
from tensor2tensor.utils.registry import *

# Adds a subsection to the registries to store specific G2G problems
registry.Registries.g2g_problems = registry.Registry("g2g_problems", validator=registry._problem_name_validator, on_set=registry._on_problem_set)
registry.Registries.g2g_hparams = registry.Registry("g2g_hparams", value_transformer=registry._hparams_value_transformer)

# Defines decorator
register_problem = lambda x: registry.register_problem(registry.Registries.g2g_problems.register(x))
register_hparams = lambda x: registry.register_hparams(registry.Registries.g2g_hparams.register(x))

# Overrides registry queries
list_g2g_problems = lambda: sorted(Registries.g2g_problems)
list_problems = list_g2g_problems
list_all_problems = list_base_problems

list_g2g_hparams = lambda: sorted(Registries.g2g_hparams)
#list_hparams = list_g2g_hparams
#list_all_hparams = registry.list_hparams



"""Install galaxy2galaxy."""

from setuptools import find_packages
from setuptools import setup
from io import open

# read the contents of the README file
with open('README.md', encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='galaxy2galaxy',
    description='Galaxy2Galaxy',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='ML4Astro Contributors',
    url='http://github.com/ml4astro/galaxy2galaxy',
    license='MIT',
    packages=find_packages(),
    package_data={'': ['*.sql'],},
    scripts=[
        'galaxy2galaxy/bin/g2g-trainer',
        'galaxy2galaxy/bin/g2g-datagen',
        'galaxy2galaxy/bin/g2g-exporter',
    ],
    install_requires=[
        'six',
        'scipy',
        'numpy',
        'astropy',
        'tensor2tensor>=1.13.4',
        'tensorflow',
        'tensorflow-hub',
        'tensorflow-datasets',
        'tensorflow-probability',
        'tensorflow-gan',
        'fits2hdf',
        'unagi>=0.1.1'
    ],
    extras_require={
        'tensorflow': ['tensorflow>=1.13.1'],
        'tensorflow_gpu': ['tensorflow-gpu>=1.13.1'],
        'tensorflow-hub': ['tensorflow-hub>=0.1.1'],
        'galsim': ['galsim'],
        'tests': [
            'absl-py',
            'pytest>=3.8.0',
            'mock',
            'pylint',
            'jupyter',
            'gsutil',
            'matplotlib',
            # Need atari extras for Travis tests, but because gym is already in
            # install_requires, pip skips the atari extras, so we instead do an
            # explicit pip install gym[atari] for the tests.
            # 'gym[atari]',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Astronomy'
    ],
    keywords='astronomy machine learning',
    use_scm_version=True,
    include_package_data=True,
    setup_requires=['setuptools_scm'],
)



"""HST/ACS Cosmos generator."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import astroimage_utils

from tensor2tensor.data_generators import generator_utils
from tensor2tensor.data_generators import image_utils
from tensor2tensor.data_generators import problem
from tensor2tensor.layers import modalities
from tensor2tensor.utils import metrics

import numpy as np
import tensorflow as tf
import os
import sys
import galsim
import argparse


class GalsimProblem(astroimage_utils.AstroImageProblem):
  """Base class for image problems generated with GalSim.

  Subclasses need only implement the `galsim_generator` function used to draw
  postage stamps with GalSim.
  """

  # START: Subclass interface
  def hparams(self, defaults, model_hparams):
    p = defaults
    p.pixel_scale = 0.03
    p.img_len = 64

  @property
  def num_bands(self):
    """Number of bands."""
    return 1

  def generator(self, data_dir, tmp_dir, dataset_split, task_id=-1):
    """
    Function to implement to generate galaxy postage stamps,
    use draw_and_encode_stamp to yield actual images from
    galsim objects
    """
    raise NotImplementedError

  def prepare_to_generate(self, data_dir, tmp_dir):
    """Prepare to generate data in parallel on different processes.

    This function is called if multiprocess_generate is True.

    Some things that might need to be done once are downloading the data
    if it is not yet downloaded, and building the vocabulary.

    Args:
      data_dir: a string
      tmp_dir: a string
    """
    pass
  # END: Subclass interface

  def example_reading_spec(self):
    """Define how data is serialized to file and read back.

    Returns:
      data_fields: A dictionary mapping data names to its feature type.
      data_items_to_decoders: A dictionary mapping data names to TF Example
         decoders, to be used when reading back TF examples from disk.
    """
    p = self.get_hparams()

    data_fields = {
        "image/encoded": tf.FixedLenFeature((), tf.string),
        "image/format": tf.FixedLenFeature((), tf.string),

        "psf/encoded": tf.FixedLenFeature((), tf.string),
        "psf/format": tf.FixedLenFeature((), tf.string),
    }


    # Adds additional attributes to be decoded as specified in the configuration
    if hasattr(p, 'attributes'):
        for k in p.attributes:
            data_fields['attrs/'+k] = tf.FixedLenFeature([], tf.float32, -1)

    data_items_to_decoders = {
        "inputs": tf.contrib.slim.tfexample_decoder.Image(
                image_key="image/encoded",
                format_key="image/format",
                channels=self.num_bands,
                shape=[p.img_len, p.img_len, self.num_bands],
                dtype=tf.float32),

        "psf": tf.contrib.slim.tfexample_decoder.Image(
                image_key="psf/encoded",
                format_key="psf/format",
                channels=self.num_bands,
                shape=[p.img_len, p.img_len, self.num_bands],
                dtype=tf.float32),
    }

    if hasattr(p, 'attributes'):
        for k in p.attributes:
            data_items_to_decoders[k] = tf.contrib.slim.tfexample_decoder.Tensor('attrs/'+k)

    return data_fields, data_items_to_decoders

  def eval_metrics(self):
    eval_metrics = [metrics.Metrics.RMSE]
    return eval_metrics

  @property
  def decode_hooks(self):
    return [image_utils.convert_predictions_to_image_summaries]

  @property
  def multiprocess_generate(self):
    """Whether to generate the data in multiple parallel processes."""
    return True

  @property
  def num_generate_tasks(self):
    """Needed if multiprocess_generate is True."""
    return self.num_train_shards + self.num_dev_shards

def _float_feature(value):
    """Returns a float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))

def _bytes_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

def draw_and_encode_stamp(gal, psf, stamp_size, pixel_scale, attributes=None):
    """
    Draws the galaxy, psf and noise power spectrum on a postage stamp and
    encodes it to be exported in a TFRecord.
    """
    # Apply the PSF
    gal = galsim.Convolve(gal, psf)

    # Draw the Fourier domain image of the galaxy
    imC = galsim.ImageCF(stamp_size, stamp_size, scale=2. *
                         np.pi / (pixel_scale * stamp_size))

    imCp = galsim.ImageCF(stamp_size, stamp_size, scale=2. *
                          np.pi / (pixel_scale * stamp_size))

    gal.drawKImage(image=imC)
    psf.drawKImage(image=imCp)

    # Keep track of the pixels with 0 value
    mask = ~(np.fft.fftshift(imC.array)[:, :(stamp_size) // 2 + 1] == 0)

    # Inverse Fourier transform of the image
    # TODO: figure out why we need 2 fftshifts....
    im = np.fft.fftshift(np.fft.ifft2(
        np.fft.fftshift(imC.array))).real.astype('float32')

    # Transform the psf array into proper format for Theano
    im_psf = np.fft.fftshift(np.fft.ifft2(
            np.fft.fftshift(imCp.array))).real.astype('float32')

    # Compute noise power spectrum
    ps = gal.noise._get_update_rootps((stamp_size, stamp_size),
                                      wcs=galsim.PixelScale(pixel_scale))

    # The following comes from correlatednoise.py
    rt2 = np.sqrt(2.)
    shape = (stamp_size, stamp_size)
    ps[0, 0] = rt2 * ps[0, 0]
    # Then make the changes necessary for even sized arrays
    if shape[1] % 2 == 0:  # x dimension even
        ps[0, shape[1] // 2] = rt2 * ps[0, shape[1] // 2]
    if shape[0] % 2 == 0:  # y dimension even
        ps[shape[0] // 2, 0] = rt2 * ps[shape[0] // 2, 0]
        # Both dimensions even
        if shape[1] % 2 == 0:
            ps[shape[0] // 2, shape[1] // 2] = rt2 * \
                ps[shape[0] // 2, shape[1] // 2]

    # Apply mask to power spectrum so that it is very large outside maxk
    ps = np.where(mask, np.log(ps**2), 10).astype('float32')

    serialized_output = {"image/encoded": [im.tostring()],
            "image/format": ["raw"],
            "psf/encoded": [im_psf.tostring()],
            "psf/format": ["raw"],
            "ps/encoded": [ps.tostring()],
            "ps/format": ["raw"]}

    # Adding the parameters provided
    if attributes is not None:
        for k in attributes:
            serialized_output['attrs/'+k] = [attributes[k]]

    return serialized_output

def maybe_download_cosmos(target_dir, sample="25.2"):
    """
    Checks for already accessible cosmos data, downloads it somewhere otherwise
    """

    import logging
    logging_level = logging.INFO

    # Setup logging to go to sys.stdout or (if requested) to an output file
    logging.basicConfig(format="%(message)s", level=logging_level, stream=sys.stdout)
    logger = logging.getLogger('galaxy2galaxy:galsim')

    try:
        catalog = galsim.COSMOSCatalog(sample=sample)
        return
    except:
        try:
            catalog = galsim.COSMOSCatalog(sample=sample, dir=target_dir)
            return
        except:
            logger.info("Couldn't access dataset, re-downloading")

    url = "https://zenodo.org/record/3242143/files/COSMOS_%s_training_sample.tar.gz"%(sample)
    file_name = os.path.basename(url)
    target = os.path.join(target_dir, file_name)
    unpack_dir = target[:-len('.tar.gz')]
    args = argparse.Namespace()
    args.quiet = True
    args.force = False
    args.verbosity = 2
    args.save = True

    # Download the tarball
    new_download, target, meta = galsim.download_cosmos.download(url, target,
                                                                 unpack_dir,
                                                                 args, logger)
    # Usually we unpack if we downloaded the tarball
    do_unpack = new_download

    # If the unpack dir is missing, then need to unpack
    if not os.path.exists(unpack_dir):
        do_unpack = True

    # But of course if there is no tarball, we can't unpack it
    if not os.path.isfile(target):
        do_unpack = False

    # Unpack the tarball
    if do_unpack:
        galsim.download_cosmos.unpack(target, target_dir, unpack_dir, meta,
                                      args, logger)

    # Usually, we remove the tarball if we unpacked it and command line doesn't specify to save it.
    do_remove = do_unpack

    # Remove the tarball
    if do_remove:
        logger.info("Removing the tarball to save space")
        os.remove(target)



""" Spectral Norm GAN """

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import tensorflow_gan as tfgan
import tensorflow_hub as hub

from tensorflow_gan.python.estimator.gan_estimator import Optimizers, get_gan_model, get_train_estimator_spec, get_eval_estimator_spec, get_predict_estimator_spec
from tensorflow_gan.python import train as tfgan_train
from tensorflow_gan.python import namedtuples
from tensorflow_gan.python.estimator.gan_estimator import SummaryType
from tensorflow.python.estimator import model_fn as model_fn_lib
from tensor2tensor.utils import hparams_lib
from tensor2tensor.utils import t2t_model
from tensor2tensor.layers import common_layers

from galaxy2galaxy.layers import spectral_ops as ops


class AbstractGAN(t2t_model.T2TModel):
  """ Base class for tf-gan based models
  """

  def generator(self, code, mode):
    raise NotImplementedError

  def discriminator(self, x, conditioning, mode):
    raise NotImplementedError

  def discriminator_loss_fn(self):
    raise NotImplementedError

  def generator_loss_fn(self):
    raise NotImplementedError

  @property
  def summaries(self):
    return [SummaryType.IMAGES]

  def sample_noise(self):
    p = self.hparams
    shape = [p.batch_size, p.bottleneck_bits]
    z = tf.random.normal(shape, name='z0', dtype=tf.float32)
    return z

  @classmethod
  def estimator_model_fn(cls,
                         hparams,
                         features,
                         labels,
                         mode,
                         config=None,
                         params=None,
                         decode_hparams=None,
                         use_tpu=False):

    if mode not in [model_fn_lib.ModeKeys.TRAIN, model_fn_lib.ModeKeys.EVAL,
                  model_fn_lib.ModeKeys.PREDICT]:
      raise ValueError('Mode not recognized: %s' % mode)

    if mode is model_fn_lib.ModeKeys.TRAIN:
      is_training = True
    else:
      is_training = False

    hparams = hparams_lib.copy_hparams(hparams)

    # Instantiate model
    data_parallelism = None
    if not use_tpu and config:
      data_parallelism = config.data_parallelism
    reuse = tf.get_variable_scope().reuse

    # Instantiate model
    self = cls(
        hparams,
        mode,
        data_parallelism=data_parallelism,
        decode_hparams=decode_hparams,
        _reuse=reuse)

    generator_inputs = self.sample_noise()
    # rename inputs for clarity
    real_data = features['inputs']
    img_shape = common_layers.shape_list(real_data)[1:4]
    real_data.set_shape([hparams.batch_size]+img_shape)

    # To satify the TFGAN API setting real data to none on predict
    if mode == tf.estimator.ModeKeys.PREDICT:
      real_data =None

    optimizers = Optimizers(tf.compat.v1.train.AdamOptimizer(
          hparams.generator_lr, hparams.beta1),
          tf.compat.v1.train.AdamOptimizer(
          hparams.discriminator_lr, hparams.beta1)
          )

    # Creates tfhub modules for both generator and discriminator
    def make_discriminator_spec():
      input_layer = tf.placeholder(tf.float32, shape=[None] + img_shape)
      disc_output = self.discriminator(input_layer, None, mode)
      hub.add_signature(inputs=input_layer, outputs=disc_output)
    disc_spec = hub.create_module_spec(make_discriminator_spec)

    def make_generator_spec():
      input_layer = tf.placeholder(tf.float32, shape=[None] + common_layers.shape_list(generator_inputs)[1:])
      gen_output = self.generator(input_layer, mode)
      hub.add_signature(inputs=input_layer, outputs=gen_output)
    gen_spec = hub.create_module_spec(make_generator_spec)

    # Create the modules
    discriminator_module = hub.Module(disc_spec, name="Discriminator_Module", trainable=True)
    generator_module = hub.Module(gen_spec, name="Generator_Module", trainable=True)

    # Wraps the modules into functions expected by TF-GAN
    generator = lambda code, mode: generator_module(code)
    discriminator =  lambda image, conditioning, mode: discriminator_module(image)

    # Make GANModel, which encapsulates the GAN model architectures.
    gan_model = get_gan_model(mode,
                              generator,
                              discriminator,
                              real_data,
                              generator_inputs,
                              add_summaries=self.summaries)

    # Make GANLoss, which encapsulates the losses.
    if mode in [tf.estimator.ModeKeys.TRAIN, tf.estimator.ModeKeys.EVAL]:
      gan_loss = tfgan_train.gan_loss(
          gan_model,
          self.generator_loss,
          self.discriminator_loss,
          add_summaries=True)

    # Make the EstimatorSpec, which incorporates the GANModel, losses, eval
    # metrics, and optimizers (if required).
    if mode == tf.estimator.ModeKeys.TRAIN:
      get_hooks_fn = tfgan_train.get_sequential_train_hooks(namedtuples.GANTrainSteps(hparams.gen_steps, hparams.disc_steps))
      estimator_spec = get_train_estimator_spec(gan_model, gan_loss, optimizers, get_hooks_fn,  is_chief=True)
    elif mode == tf.estimator.ModeKeys.EVAL:
      estimator_spec = get_eval_estimator_spec(gan_model, gan_loss)
    else:  # tf.estimator.ModeKeys.PREDICT
      # Register hub modules for export
      hub.register_module_for_export(generator_module, "generator")
      hub.register_module_for_export(discriminator_module, "discriminator")
      estimator_spec = get_predict_estimator_spec(gan_model)
    return estimator_spec
